/*
 * Portions of this software are copyrighted by the 
 * 
 *     University of South Carolina,
 *     Sun Microsystems, Inc., 
 * 
 * and other parties. 
 * 
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
 * DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
 * IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
 * NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 */


package edu.sc.seis.TauP;

import java.util.*;
import java.util.zip.*;
import java.io.*;

/** convenience class for loading properties.
  *
  * @version 1.0  Tue Feb  2 20:48:34 GMT 1999
  * @author H. Philip Crotwell
  *
  */
public class PropertyLoader {

	protected static String jarFileName = "taup.jar";

	protected static String defaultPropFileName = "defaultProps";

	protected static String userPropFileName = ".taup";

		/** loads the properties from a file. First the default properties
		 *  are loaded from the distribution jar file, then the users
		 *  properties are loaded, overwriting the default values. 
		 *  This uses ".taup" in the users home directory, followed by 
		 *  ".taup" in the current directory if it can be found. 
		 *  If neither can be found then the default Properties object 
		 *  is returned unmodified.
		 *
		 *  A special case is made for the taup.model.path property. If it
		 *  is defined in the system properties, then the system version
		 *  is prepended to the users version. This allows for setting
		 *  system wide search paths on UNIX via an environment variable, 
		 *  which is transformed into a property by the sh scripts, while
		 *  still allowing individual users as well as non-UNIX systems 
		 *  to customize the search path.  */
	public static Properties load() throws IOException {

		Properties defaultProps = new Properties();

			// load default properties
		try {
			String classPath = System.getProperty("java.class.path");
			String pathEntry = "";
			int offset = 0;
			int pathSepIndex;
			File jarFile;

				// loop over each entry in the CLASSPATH, looking for taup.jar
			while (offset < classPath.length()) {
				pathSepIndex = classPath.indexOf(File.pathSeparatorChar, offset);
				if (pathSepIndex != -1) {
					pathEntry = classPath.substring(offset,pathSepIndex);
					offset = pathSepIndex+1;
				} else {
					pathEntry = classPath.substring(offset);
					offset = classPath.length();
				}
				if (pathEntry.endsWith(jarFileName)) {
					jarFile = new File(pathEntry);
					if (jarFile.exists() && jarFile.isFile() &&
					jarFile.getName().equals(jarFileName) && jarFile.canRead()) {
						ZipFile zippy = new ZipFile(jarFile);
						ZipEntry zipEntry = zippy.getEntry(defaultPropFileName);
						if (zipEntry != null) {
							defaultProps.load( zippy.getInputStream( zipEntry));
							zippy.close();

								//we've found the path to the jar, so exit the while
							offset = classPath.length()+1;
						}
					}
				}
			}
		} catch (FileNotFoundException e) {
			// can't find defaults, so we'll just have to use an empty
			// properties object
		}
			

			// create program properties with default
		Properties applicationProps = new Properties(defaultProps);

			// append/overwrite with user's directory .taup
		try {
			applicationProps.load( new FileInputStream(
				System.getProperty("user.home")+
				System.getProperty("file.separator")+".taup"));
		} catch (FileNotFoundException ee) {
			// file doesn't exist, so go on
		}

			// append/overwrite with current directory .taup
		try {
		      /* Check for .taup in the current directory. */
	      applicationProps.load( new FileInputStream(
				System.getProperty("user.dir")+
				System.getProperty("file.separator")+".taup"));
		} catch (FileNotFoundException e) {
			// file doesn't exist, so go on
		}

			// check for taup.model.path in system properties
		String taupPath = "taup.model.path";
		Properties sysProps = System.getProperties();
		if (sysProps.containsKey(taupPath)) {
			if (applicationProps.containsKey(taupPath)) {
				applicationProps.put(taupPath, sysProps.getProperty(taupPath) +
					sysProps.getProperty("path.separator") +
					applicationProps.getProperty(taupPath));
			} else {
				applicationProps.put(taupPath, sysProps.getProperty(taupPath));
			}
		}

		return applicationProps;
	}

		/** writes the current system properties out to the file given. */
	public static void save(Properties props) throws IOException {
		save(props, ".taup");
	}

		/** writes the current system properties out to the file given. */
	public static void save(Properties props, String filename) 
	throws IOException {
		FileOutputStream propFile = new FileOutputStream(filename);
		props.save(propFile, "---Properties for the TauP toolkit---");
		propFile.close();
	}

	public static void main(String[] args) {
		try {
			Properties props = PropertyLoader.load();
			props.put("Key", "Value and another value");
			save(props, "testProperties");
		} catch (IOException e) {
			System.out.println("Caught IOException: "+e.getMessage());
		}
	}
}
